'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.convertPathMapToRelativePaths = convertPathMapToRelativePaths;
exports.getPathMapWithLinkFilesData = getPathMapWithLinkFilesData;

var _ramda = require('ramda');

var _ramda2 = _interopRequireDefault(_ramda);

var _generateTreeMadge = require('./generate-tree-madge');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * PathMap is used to get the ImportSpecifiers from dependency-tree library
 */

/**
 * Path Map is the extra data about the files and the dependencies, such as ImportSpecifiers and
 * custom-resolve-modules used.
 * The data is retrieved by dependency-tree which collects them from the various detectives.
 * It is used to update the final tree.
 */
function convertPathMapToRelativePaths(pathMap, baseDir) {
  var pathCache = {};
  return pathMap.map(function (file) {
    var newFile = _ramda2.default.clone(file);
    newFile.file = (0, _generateTreeMadge.processPath)(file.file, pathCache, baseDir);
    newFile.dependencies = file.dependencies.map(function (dependency) {
      var newDependency = _ramda2.default.clone(dependency);
      newDependency.resolvedDep = (0, _generateTreeMadge.processPath)(dependency.resolvedDep, pathCache, baseDir);
      return newDependency;
    });
    return newFile;
  });
}

/**
 * if a resolvedDep of a specifier has its own importSpecifiers and one of them is the same of the
 * given specifier, then, the realDep is not the resolvedDep and might be the resolvedDep of the
 * resolvedDep. It needs to be recursive because it might have multiple link files.
 *
 * it is easier to understand with an example:
 * bar/foo.js requires an index file => utils/index.js, => `import { isString } from '../utils';`
 * which requires another index file: utils/is-string/index.js, => `export { default as isString } from './is-string';`
 * which requires a real file utils/is-string/is-string.js => `export default function isString() { };`
 *
 * according to the files, the dependency of bar/foo is utils/index.js.
 * however, we suspect that utils/index.js is only a link file so we try to find the realDep.
 *
 * round 1:
 * utils/is-string/index.js is a first candidate because it has isString in one of the dependencies
 * importSpecifiers. however, since it has dependencies itself, we might find more candidates.
 * continue to round 2 to find out.
 *
 * round 2:
 * isString specifier of utils/is-string/index.js is connected to utils/is-string/is-string.js
 * which introduce this file as the second candidate. this file doesn't have any dependency,
 * therefore it returns this file as the final realDep.
 */
function findTheRealDepRecursive(pathMap, depPathMap, specifier, lastRealDep) {
  var realDep = depPathMap.dependencies.find(function (dep) {
    if (!dep.importSpecifiers) return false;
    return dep.importSpecifiers.find(function (depSpecifier) {
      return depSpecifier.name === specifier.name;
    });
  });
  if (!realDep) return lastRealDep;
  var realDepPathMap = pathMap.find(function (file) {
    return file.file === realDep.resolvedDep;
  });
  if (!realDepPathMap || !realDepPathMap.dependencies.length) return realDep;
  return findTheRealDepRecursive(pathMap, realDepPathMap, specifier, realDep);
}

/**
 * if a dependency file is in fact a link file, get its real dependencies.
 */
function getDependenciesFromLinkFileIfExists(dependency, pathMap) {
  var dependencyPathMap = pathMap.find(function (file) {
    return file.file === dependency.resolvedDep;
  });
  if (!dependencyPathMap || !dependencyPathMap.dependencies.length || !dependency.importSpecifiers || !dependency.importSpecifiers.length) {
    return null;
  }

  var dependencies = dependency.importSpecifiers.map(function (specifier) {
    var realDep = findTheRealDepRecursive(pathMap, dependencyPathMap, specifier);
    // $FlowFixMe
    if (!realDep) return null;
    // $FlowFixMe importSpecifiers do exist
    var depImportSpecifier = realDep.importSpecifiers.find(function (depSpecifier) {
      return depSpecifier.name === specifier.name;
    });
    var importSpecifier = {
      mainFile: specifier,
      linkFile: depImportSpecifier
    };
    return { file: realDep.resolvedDep, importSpecifier };
  });

  if (dependencies.some(function (dep) {
    return !dep;
  })) {
    // at least one specifier doesn't have "realDep", meaning it doesn't use link-file
    return null;
  }
  var linkFiles = [];
  dependencies.forEach(function (dep) {
    var existingFile = linkFiles.find(function (linkFile) {
      return linkFile.file === dep.file;
    });
    if (existingFile) {
      existingFile.importSpecifiers.push(dep.importSpecifier);
    } else {
      linkFiles.push({ file: dep.file, importSpecifiers: [dep.importSpecifier] });
    }
  });
  return linkFiles;
}

/**
 * mark dependencies that are link-files as such. Also, add the data of the real dependencies
 */
function getPathMapWithLinkFilesData(pathMap) {
  if (!Array.isArray(pathMap)) throw new TypeError('pathMap must be an array');
  var updateDependencyWithLinkData = function updateDependencyWithLinkData(dependency) {
    var dependenciesFromLinkFiles = getDependenciesFromLinkFileIfExists(dependency, pathMap);
    if (dependenciesFromLinkFiles && dependenciesFromLinkFiles.length) {
      // it is a link file
      dependency.linkFile = true;
      dependency.realDependencies = dependenciesFromLinkFiles;
    }
  };
  pathMap.forEach(function (file) {
    file.dependencies.forEach(function (dependency) {
      updateDependencyWithLinkData(dependency);
    });
  });

  return pathMap;
}
//# sourceMappingURL=path-map.js.map